library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;

entity DESIGN_TOP is
    port (
    	DA         : in  STD_LOGIC_VECTOR (11 downto 0);   -- from Cavity Field ADC
    	DAOV       : in  STD_LOGIC;
    	DB         : in  STD_LOGIC_VECTOR (11 downto 0);   -- from Cavity Field ADC
    	DBOV       : in  STD_LOGIC;
    	DC         : in  STD_LOGIC_VECTOR (11 downto 0);   -- from Cavity Field ADC
    	DCOV       : in  STD_LOGIC;
    	DD         : in  STD_LOGIC_VECTOR (11 downto 0);   -- from Cavity Field ADC
    	DDOV       : in  STD_LOGIC;
    	SYNC	   : in  STD_LOGIC;    -- external "10 MHz" sync 
    	GLOBAL_RST : in  STD_LOGIC;
    	CLK        : in  STD_LOGIC;
    	P_Down	   : out STD_LOGIC;
    	RF_ON      : in  STD_LOGIC;    -- external RF gate pulse
    	RF_KILL    : out STD_LOGIC;   -- we can disable the RF from here
    	
    	-- host interface
    	PD	 : inout STD_LOGIC_VECTOR (15 downto 0);
    	ALE_host : in  STD_LOGIC;
    	WE_host  : in  STD_LOGIC;
    	RD_host  : in  STD_LOGIC;
    	CS0_host : in  STD_LOGIC;
    	RDY_host : out STD_LOGIC;
    	CK_host  : in STD_LOGIC;
    	URST     : in  STD_LOGIC;         -- unused, I'm not even sure this is an input
    	INT_host : out STD_LOGIC;
    	
    	-- DAC control pins
    	DE  	 : out UNSIGNED (11 downto 0);   -- ERROR OUT to DAC
    	OCLK 	 : OUT std_logic;
	OSEL 	 : OUT std_logic;
	--IQRST  : OUT std_logic;
	SLEEP 	 : OUT std_logic;

	-- Serial port to external ADC/DAC
	SP_DOUT	 : in  std_logic;	-- to FPGA sport from ADC/DAC
	SP_DIN   : out std_logic;	-- from FPGA sport to ADC/DAC
	SP_SCLK	 : out std_logic;
	SP_CS1	 : out std_logic;	-- CS1 on schematic
	SP_CS2	 : out std_logic;	-- CS2 on schematic
	SP_CS3	 : out std_logic;	-- CS3 on schematic
	SP_CS4	 : out std_logic;	-- CS4 on schematic
	
	-- Test points
	TEST_1   : out std_logic;
	TEST_2   : out std_logic;
	TEST_3   : out std_logic;
	TEST_4   : out std_logic
    );
end DESIGN_TOP;

architecture DESIGN_TOP_arch of DESIGN_TOP is
-- component declarations

component ramdp1024x12 is
  port (
	ADDRa: IN std_logic_vector(9 downto 0);
	RSTa : IN std_logic;
	ENa  : IN std_logic;
	WEa  : IN std_logic;
	CLKa : IN std_logic;
 	DIa  : IN  std_logic_vector(11 downto 0);
	DOa  : OUT std_logic_vector(11 downto 0);

	ADDRb: IN std_logic_vector(9 downto 0);
	RSTb : IN std_logic;
	ENb  : IN std_logic;
	WEb  : IN std_logic;
	CLKb : IN std_logic;
 	DIb  : IN  std_logic_vector(11 downto 0);
	DOb  : OUT std_logic_vector(11 downto 0)
  );
end component;

component ramdp1024x32 is
  port (
	ADDRa: IN std_logic_vector(9 downto 0);
	RSTa : IN std_logic;
	ENa  : IN std_logic;
	WEa  : IN std_logic;
	CLKa : IN std_logic;
 	DIa  : IN  std_logic_vector(31 downto 0);
	DOa  : OUT std_logic_vector(31 downto 0);

	ADDRb: IN std_logic_vector(9 downto 0);
	RSTb : IN std_logic;
	ENb  : IN std_logic;
	WEb  : IN std_logic;
	CLKb : IN std_logic;
 	DIb  : IN  std_logic_vector(31 downto 0);
	DOb  : OUT std_logic_vector(31 downto 0)
  );
end component;

component RAMB4_S8_S8 is
    port (
  	ENa   : in STD_LOGIC;
  	DIa   : in STD_LOGIC_VECTOR (7 downto 0);
  	ADDRa : in STD_LOGIC_VECTOR (8 downto 0);
  	WEa   : in STD_LOGIC;
  	RSTa  : in STD_LOGIC;
  	CLKa  : in STD_LOGIC;
  	DOa   : out STD_LOGIC_VECTOR (7 downto 0);

	ENb   : in STD_LOGIC;  	
  	DIb   : in STD_LOGIC_VECTOR (7 downto 0);
  	ADDRb : in STD_LOGIC_VECTOR (8 downto 0);
  	WEb   : in STD_LOGIC;
  	RSTb  : in STD_LOGIC;
  	CLKb  : in STD_LOGIC;
  	DOb   : out STD_LOGIC_VECTOR (7 downto 0)
    );
end component;

component FDBK_LOOP is
    port (
    	M        : in SIGNED (11 downto 0);
    	SET_I    : in SIGNED (13 downto 0);
    	SET_Q    : in SIGNED (13 downto 0);
    	KCM_load_data : in SIGNED (11 downto 0);
    	KCM_load_real : in STD_LOGIC;
    	KCM_load_imag : in STD_LOGIC;
    	KCM_load_intg : in STD_LOGIC;
    	KCM_load_busy : out STD_LOGIC;
    	t_mod_4  : in STD_LOGIC_VECTOR (1 downto 0);
    	feedback_enable  : in STD_LOGIC;
    	integrate_enable : in STD_LOGIC;
    	feedforward_data: in  STD_LOGIC_VECTOR(7 downto 0);
    	RESET    : in STD_LOGIC;
    	CLK      : in STD_LOGIC;
    	FDBK_ERR : out SIGNED (11 downto 0)    
    );
end component;

component sport is
    port (
  	write: in std_logic;
  	addr : in std_logic_vector (1 downto 0);
  	idata: in std_logic_vector (15 downto 0);
  	din  : in std_logic;
   	clk  : in std_logic;
   	dout : out std_logic;
  	odata: out std_logic_vector (15 downto 0);
  	sclk : out std_logic;
  	cs0  : out std_logic;
  	cs1  : out std_logic;
  	cs2  : out std_logic;
  	cs3  : out std_logic
    );
end component;

component rf_timer is
    port (
	ck:              in std_logic;  -- 40 MHz
	quad_enable:     in std_logic;  -- 10 MHz
	halt:            in std_logic;
	rf:              in std_logic;
	ignore_rf_off:   in std_logic;
	self_retrigger:  in std_logic;
	write_pulse:     in std_logic;
	host_address:    in std_logic_vector(2 downto 0);
	host_data:       in std_logic_vector(15 downto 0);
	host_read:       out std_logic_vector(15 downto 0);
	feedback_run:    out std_logic;
	integrate_run:   out std_logic;
	feedforward_start: out std_logic;
	trace_enable:    out std_logic;
	adc_power:       out std_logic;
	host_interrupt:  out std_logic;
	rf_gate:         out std_logic;
	trigger_skipped: out std_logic
    );
end component;

component BUFGDLL
    port (
	I	: in std_logic;
	O	: out std_logic
    );
end component;

component IBUFG 
    port (
	I: IN std_logic;
	O: OUT std_logic
    );
end component;

component BUFG
 	port (
 	I: IN std_logic;
 	O: OUT std_logic);
end component;
-- end of component declarations

-- internal signal declarations
signal CLK_40MHz : std_logic;
signal CLK_buf   : std_logic;

signal SET_I    :   signed (13 downto 0);
signal SET_Q    :   signed (13 downto 0);
signal M_signed :   signed (11 downto 0);
signal ZERO_12  : std_logic_vector (11 downto 0);
signal ZERO_32  : std_logic_vector (31 downto 0);
signal KCM_load_data : signed (11 downto 0);
signal KCM_load_real : std_logic;
signal KCM_load_imag : std_logic;
signal KCM_load_intg : std_logic;
signal KCM_load_busy : std_logic;


signal feedback_output  : signed (11 downto 0);
signal decay_trace_dout : std_logic_vector (11 downto 0);
signal wide_trace_dout  : std_logic_vector (31 downto 0);
signal wide_trace_din   : std_logic_vector (31 downto 0);
signal trace_address    : std_logic_vector (15 downto 0);
signal wide_address     : std_logic_vector (9 downto 0);
signal wide_trace_config: std_logic_vector (5 downto 0);
signal address_at_pulse_end : std_logic_vector (15 downto 0);
signal trace_select, feedback_select, integrate_select  : std_logic;  -- set by software
signal trace_run,    feedback_run,    integrate_run     : std_logic;  -- output of timer
signal trace_enable, feedback_enable, integrate_enable  : std_logic;  -- select & run, controls data path
signal test_point_enable: std_logic;
signal LO               : STD_LOGIC;
SIGNAL HI               : STD_LOGIC;
signal RF_ON_sync       : std_logic;
signal RF_ON_hold	: std_logic;
signal external_trigger	: std_logic;
signal SYNC_delayed	: std_logic;
signal SYNC_sync        : std_logic;
signal SYNC_last	: std_logic;
signal sync_count	: std_logic_vector (4 downto 0);
signal quad_sync	: std_logic;
signal error_clear_cmd  : std_logic;
signal sync_error	: std_logic;
signal sync_missing	: std_logic;
signal ext_trigger_skipped : std_logic;
signal adc_power        : std_logic;
signal rf_gate          : std_logic;

signal da_latch, db_latch, dc_latch, dd_latch : std_logic_vector (11 downto 0);
signal trace_data : std_logic_vector(11 downto 0);
signal trace_sel : std_logic_vector(1 downto 0);
signal daov_latch, dbov_latch, dcov_latch, ddov_latch : std_logic;

-- for the missing clock detect logic, processes detect_40MHz_gray and detect_40MHz.
signal gray_40MHz    : std_logic_vector(1 downto 0);
signal gray_sample   : std_logic_vector(1 downto 0);
signal gray_previous : std_logic_vector(1 downto 0);
signal clk40_missing : std_logic;
signal ext_trigger_skipped_latch : std_logic;

signal feedforward_we        : std_logic;
signal feedforward_start     : std_logic;
signal feedforward_counter   : std_logic_vector (10 downto 0);
signal feedforward_rf_addr   : std_logic_vector (8 downto 0);
signal feedforward_rf_dout   : std_logic_vector (7 downto 0);
signal feedforward_host_dout : std_logic_vector (7 downto 0);
signal zero_byte             : std_logic_vector (7 downto 0);

signal hCLK_buf       : std_logic;
signal CLK_host       : std_logic;
signal ADDR_reg       : std_logic_vector (13 downto 0);
signal pulse_start_en_buffer : std_logic;
signal status         : std_logic_vector (7 downto 0);
signal errors         : std_logic_vector (7 downto 0);

signal host_write_cycle : boolean;
signal KCM_load_start   : boolean;
signal KCM_load_sel     : std_logic_vector(1 downto 0);
signal KCM_load_cmd     : std_logic;
signal KCM_load_delayed : std_logic;

-- configured by host with writes to 0x0028
signal pulse_start_en        : std_logic;
signal phase_sense_adjust    : std_logic;
signal integrate_err_gate    : std_logic;
signal continuous_feedforward: std_logic;
signal ignore_rf_off         : std_logic;
signal self_retrigger        : std_logic;
signal halt                  : std_logic;
signal iq_hand_flip          : std_logic;

signal host_sport_write : std_logic;
signal host_sport_odata : std_logic_vector (15 downto 0);
signal host_timer_write : std_logic;
signal host_timer_odata : std_logic_vector (15 downto 0);

signal hack1 : std_logic;
signal hack2 : std_logic;

-- end of signal declarations

begin
LO <= '0';
HI <= '1';
zero_byte <= "00000000";

-- static, currently unused output to host:
RDY_host  <= '1';

-- not yet used feature to kill the RF if we detect something
-- wrong from our end.
RF_KILL <= '0';

-- buffer the clocks
u0: IBUFG port map (I => CLK,        O => CLK_buf   );
u1:  BUFG port map (I => CLK_buf,    O => CLK_40MHz ); 
u3:  BUFG port map (I => CK_host,    O => CLK_host  );

-- ideally this latch would use the IOB flip-flop
dac_output_latch: process (CLK_40MHz) begin
    if (CLK_40MHz'event and CLK_40MHz = '1') then
        DE <= unsigned(not feedback_output(11) & feedback_output(10 downto 0));
    end if;
end process dac_output_latch;

M_signed <= SIGNED(not DC(11) & DC(10 downto 0));
ZERO_12 <= "000000000000";
ZERO_32 <= "00000000000000000000000000000000";

trace_enable     <= trace_run     and trace_select;
feedback_enable  <= feedback_run  and feedback_select;
integrate_enable <= integrate_run and integrate_select;

-- design components instantiation
decay_trace: ramdp1024x12 port map
      (
  	ADDRa => trace_address(9 downto 0),
  	RSTa  => LO,
  	ENa   => HI,
  	WEa   => trace_enable,
  	CLKa  => CLK_40MHz,  	
  	DIa   => trace_data,
  	--DOa => ,
  	
  	-- read from host
  	ADDRb => ADDR_reg(9 downto 0),
  	RSTb  => LO,
  	ENb   => HI,
  	WEb   => LO,			-- host write not allowed
  	CLKb  => CLK_host,
  	DIb   => ZERO_12,			-- not used
  	DOb   => decay_TRACE_dout 
       );

wide_trace_din <= da_latch(11 downto 1) & db_latch(11 downto 1) & dc_latch(11 downto 2);
wide_trace: ramdp1024x32 port map
      (
  	ADDRa => wide_address,
  	RSTa  => LO,
  	ENa   => HI,
  	WEa   => trace_enable,
  	CLKa  => CLK_40MHz,  	
  	DIa   => wide_trace_din,
  	--DOa => ,
  	
  	-- read from host
  	ADDRb => ADDR_reg(9 downto 0),
  	RSTb  => LO,
  	ENb   => HI,
  	WEb   => LO,			-- host write not allowed
  	CLKb  => CLK_host,
  	DIb   => ZERO_32,			-- not used
  	DOb   => wide_TRACE_dout 
       );

feedforward_table: RAMB4_S8_S8 port map
      (
        ENa   => HI,
  	DIa   => zero_byte,			-- not used
  	ADDRa => feedforward_rf_addr,
  	WEa   => LO,				-- rf reads only
  	RSTa  => GLOBAL_RST,
  	CLKa  => CLK_40MHz,
  	DOa   => feedforward_rf_dout,
  	
  	-- read from host
  	ENb   => HI,
  	DIb   => PD(7 downto 0),
  	ADDRb => ADDR_reg(8 downto 0),
  	WEb   => feedforward_we,
  	RSTb  => GLOBAL_RST,
  	CLKb  => CLK_host,
  	DOb   => feedforward_host_dout 
       );


fdbk: FDBK_LOOP port map
       (
    	M       => M_signed,
    	SET_I   => SET_I,
    	SET_Q   => SET_Q,
    	KCM_load_data => KCM_load_data,
    	KCM_load_real => KCM_load_real,
    	KCM_load_imag => KCM_load_imag,
    	KCM_load_intg => KCM_load_intg,
    	KCM_load_busy => KCM_load_busy,
    	t_mod_4 => sync_count(1 downto 0),
    	feedback_enable => feedback_enable,
    	integrate_enable => integrate_enable,
    	feedforward_data => feedforward_rf_dout, 
    	RESET   => GLOBAL_RST, 
    	CLK     => CLK_40MHz,
    	FDBK_ERR=> feedback_output 
       );

-- Note the reversal of din/dout at this instantiation.
-- The sport implementation (naturally) drives its dout
-- and reads its din.  The pins on the FPGA in the circuit
-- attach to wires in the system that connect to the
-- ADC and DAC din and dout, and by reversing din/dout
-- here, we keep those wires labelled consistently
-- even on the FPGA.  So the FPGA drives the SP_din pin,
-- and reads the SP_dout pin.
serial: sport port map
       (
        write => host_sport_write,
  	addr  => ADDR_reg(1 downto 0),
  	idata => PD,
  	din   => SP_dout,	
   	clk   => CLK_host,
   	dout  => SP_din,	
  	odata => host_sport_odata,
  	sclk  => SP_sclk,
  	-- set the mapping from logical address to physical chip selects
  	cs0   => SP_cs1,
  	cs1   => SP_cs2,
  	cs2   => SP_cs3,
  	cs3   => SP_cs4
  	);

-- TIMING CONTROL LOGIC
rf_timer1: rf_timer port map
      (
	ck               => CLK_40MHz,
	quad_enable      => quad_sync,
	halt             => halt,
	rf               => RF_ON_sync,
	ignore_rf_off    => ignore_rf_off,
	self_retrigger   => self_retrigger,
	write_pulse      => host_timer_write,
	host_address     => ADDR_reg(2 downto 0),
	host_data        => PD,
	host_read        => host_timer_odata,
	feedback_run     => feedback_run,
	integrate_run    => integrate_run,
	feedforward_start => feedforward_start,
	trace_enable     => trace_run,
	adc_power        => adc_power,
	host_interrupt   => INT_host,
	rf_gate          => rf_gate,   -- XXX unused
	trigger_skipped  => ext_trigger_skipped
	);

P_down <= not adc_power;

-- Feature list:
--   An incoming 2.5, 5, or 10 MHz signal on SYNC will lock the internal
--   phase counter sync_count.  If the incoming sync signal is missing,
--   the counter will overflow and the sync_missing flag will be set.
--   If the incoming sync signal does not stay in phase with CLK_40MHz/4,
--   sync_error will be set.  This might also happen if SYNC rises very
--   near the rising edge of CLK_40MHz, causing single cycle jitter in
--   SYNC_sync.  If that happens, the software has to set phase_sense_adjust,
--   which makes this logic listen to a copy of SYNC that is latched on the
--   falling edge of CLK_40MHz.
--
--   SYNC is latched on both the rising and falling edge of CLK_40MHz, and
--   it is not possible to put both latches in the Virtex IOB.  So timing
--   is what it is, fortunately there's almost no logic to get in the way.
--   The system just has to route from the IOB, through one level of logic,
--   to the two latches with much less than 12 ns skew.
--
--   The incoming RF_ON signal is synchronized to 40 MHz.  In rf_timer, a
--   rising edge is noticed and creates a one-clock-long external_trigger pulse.
--   This trigger is synchronized to the internal 10 MHz quadrature clock.
--   The RF_ON signal must be longer than 100ns for this logic to work,
--   not a problem since it's expected to be 100us to 1000us long.
phase_shift: process (CLK_40MHz)
begin
  if (CLK_40MHz'event and CLK_40MHz = '0') then
    SYNC_delayed <= SYNC;
  end if;
end process phase_shift;

phase_sync: process (CLK_40MHz)
begin
  if (CLK_40MHz'event and CLK_40MHz = '1') then
    RF_ON_sync <= RF_ON;                 -- single point synchronization of external signal
    -- SYNC is asynchronous.  It passes through simple asynchonous logic,
    -- and is sampled here in a single latch.  If phase_sense_adjust is set,
    -- SYNC is ignored here, and the falling-edge sampled version SYNC_delayed
    -- is used instead.
    SYNC_sync <= (phase_sense_adjust and SYNC_delayed) or (not phase_sense_adjust and SYNC);
    SYNC_last <= SYNC_sync;
    if (SYNC_sync = '1' and SYNC_last = '0') then
      sync_count <= "00000";
      if (not sync_count(1 downto 0) = "11") then  
        sync_error <= '1';
      elsif(error_clear_cmd = '1') then  -- error_clear_cmd is asynchronous, but it doesn't matter much
        sync_error <= '0';
      end if;
    else
      sync_count <= sync_count + 1;
    end if;
    if (sync_count(1 downto 0) = "11") then
      quad_sync <= '1';
    else
      quad_sync <= '0';
    end if;
    if (sync_count(4) = '1') then
      sync_missing <= '1';
    elsif (error_clear_cmd = '1') then
      sync_missing <= '0';
    end if;
  end if;
end process phase_sync;

-- these two processes provide a "clock detect" feature.
-- If the 40 MHz goes away, or is even erratic, an error bit is raised.
-- If the 25 MHz CPU clock goes away, there's nobody to tell :-(
-- Strategy for crossing the clock domain: use a 2-bit Gray code
-- counter clocked at 40 MHz, which can be sampled at 25 MHz.
-- The counter will always have advanced one or two 25 ns cycles
-- between 40 ns clocks.
detect_40MHz_gray: process (CLK_40MHz)
begin
  if (CLK_40MHz'event and CLK_40MHz = '1') then
    gray_40MHz <= gray_40MHz(0) & not gray_40MHz(1);
  end if;
end process detect_40MHz_gray;
detect_40MHz: process (CLK_host)
begin
  if (CLK_host'event and CLK_host = '1') then
    gray_sample <= gray_40MHz;    -- cross clock domains *here*
    gray_previous <= gray_sample;
    if (gray_previous = gray_sample) then
      clk40_missing <= '1';
    elsif (error_clear_cmd = '1') then
      clk40_missing <= '0';
    end if;
  end if;
end process detect_40MHz;


address: process (CLK_40MHz)
begin 
  if (CLK_40MHz'event and CLK_40MHz='1') then
    if (trace_enable = '1') then
      address_at_pulse_end <= trace_address;
      trace_address <= trace_address + 1;
    else
      trace_address <= (others=>'0');
    end if;
    if (ext_trigger_skipped = '1') then
      ext_trigger_skipped_latch <= '1';
    elsif (error_clear_cmd = '1') then
      ext_trigger_skipped_latch <= '0';
    end if;
    da_latch <= da;
    db_latch <= db;
    dc_latch <= dc;
    dd_latch <= dd;
    if (trace_sel(1) = '0') then
        if (trace_sel(0) = '0') then
            trace_data <= da;
        else
            trace_data <= db;
        end if;
    else
        if (trace_sel(0) = '0') then
            trace_data <= dc;
        else
            trace_data <= dd;
        end if;
    end if;
    if (trace_enable='1' and daov='1') then
       daov_latch <= '1';
    elsif (error_clear_cmd='1') then
       daov_latch <= '0';
    end if;
    if (trace_enable='1' and dbov='1') then
       dbov_latch <= '1';
    elsif (error_clear_cmd='1') then
       dbov_latch <= '0';
    end if;
    if (trace_enable='1' and dcov='1') then
       dcov_latch <= '1';
    elsif (error_clear_cmd='1') then
       dcov_latch <= '0';
    end if;
    if (trace_enable='1' and ddov='1') then
       ddov_latch <= '1';
    elsif (error_clear_cmd='1') then
       ddov_latch <= '0';
    end if;
 end if;
end process address;

-- 11 bit counter that is used to generate the feedforward address.
-- ignore bits 1 and 2 to generate the address for the 512 element
-- feedforward table.  Host needs to make sure the value at the
-- first address is zero, since we sit there when the system is
-- idle.  
feedforward_addr: process (CLK_40MHz)
begin
    if (CLK_40MHz'event and CLK_40MHz = '1') then
	if ( (continuous_feedforward = '1') or (feedforward_start='1')
	     or not (feedforward_counter="00000000000")) then
		feedforward_counter <= feedforward_counter + 1;
	else
		feedforward_counter <= (others=>'0');
	end if;
    end if;
end process feedforward_addr;
feedforward_rf_addr <= feedforward_counter(10 downto 3) & feedforward_counter(0);
wide_address(0) <= trace_address(0);
wide_address(1) <= trace_address(1);
wide_address(2) <= trace_address(2) when wide_trace_config(0)='0' else trace_address(10);
wide_address(3) <= trace_address(3) when wide_trace_config(1)='0' else trace_address(11);
wide_address(4) <= trace_address(4) when wide_trace_config(2)='0' else trace_address(12);
wide_address(5) <= trace_address(5) when wide_trace_config(3)='0' else trace_address(13);
wide_address(6) <= trace_address(6) when wide_trace_config(4)='0' else trace_address(14);
wide_address(7) <= trace_address(7) when wide_trace_config(5)='0' else trace_address(15);
wide_address(8) <= trace_address(8);
wide_address(9) <= trace_address(9);
--wide_address <= trace_address(9 downto 0);

test_point: process (CLK_40MHz) begin
    if (CLK_40MHz'event and CLK_40MHz = '1') then
        if (test_point_enable ='1') then
            TEST_1 <= feedforward_start;
            TEST_2 <= integrate_run;
            TEST_3 <= feedforward_counter(1);
            TEST_4 <= '1';
        else
    	    TEST_1 <= '0';
    	    TEST_2 <= '0';
    	    TEST_3 <= '0';
    	    TEST_4 <= '0';
    	end if;
    end if;
end process test_point;


KCM_load_start <= (KCM_load_cmd='1') and (KCM_load_delayed='0') and (KCM_load_busy='0');

KCM_write_control: process (CLK_host) begin
    if (CLK_host'event and CLK_host = '1') then
        KCM_load_delayed <= KCM_load_cmd;
        if (KCM_load_start and (KCM_load_sel="00")) then
            KCM_load_real <= '1';
        elsif (KCM_load_busy='1') then
            KCM_load_real <= '0';
        end if;
        if (KCM_load_start and (KCM_load_sel="01")) then
            KCM_load_imag <= '1';
        elsif (KCM_load_busy='1') then
            KCM_load_imag <= '0';
        end if;
        if (KCM_load_start and (KCM_load_sel="10")) then
            KCM_load_intg <= '1';
        elsif (KCM_load_busy='1') then
            KCM_load_intg <= '0';
        end if;
    end if;
end process KCM_write_control;

host_write_cycle <= (WE_host = '0') and (CS0_host = '0');

-- host interface control
host_IF: process (CLK_host) begin
    if (CLK_host'event and CLK_host = '1') then
        if (ALE_host = '1' and CS0_host = '0') then
             -- ignore lower two bits, we don't support byte or halfword operations on
             -- this virtual address space.
             ADDR_reg <= PD(15 downto 2);
        end if;
        if (WE_host = '0' and CS0_host = '0') then
            if (ADDR_reg(13 downto 4) = "0000000000") then
                if (ADDR_reg(3 downto 0) = "1000") then -- 0x0020
                    SET_I <= SIGNED(PD(15 downto 2));
                end if;
                if (ADDR_reg(3 downto 0) = "1001") then -- 0x0024
                    SET_Q <= SIGNED(PD(15 downto 2));
                end if;
                if (ADDR_reg(3 downto 0) = "1010") then -- 0x0028
                    trace_select           <= PD(0);
                    phase_sense_adjust     <= PD(1);
                 -- integrate_err_gate     <= PD(2);
                    continuous_feedforward <= PD(3);
                    ignore_rf_off          <= PD(4);
                    self_retrigger         <= PD(5);
	            halt                   <= PD(6);
	            trace_sel              <= PD(8 downto 7);
	            feedback_select        <= PD(9);
                    integrate_select       <= PD(10);
	            test_point_enable      <= PD(11);
	            iq_hand_flip           <= PD(12);
                end if;
                if (ADDR_reg(3 downto 0) = "1011") then -- 0x002c
                    wide_trace_config      <= PD(5 downto 0);
                end if;
                if ( ADDR_reg(3 downto 2) = "11") then -- 0x0030-0x003c  KCM-xxx
                    KCM_load_data <= SIGNED(PD(15 downto 4));
                    KCM_load_sel  <= ADDR_reg(1 downto 0);
                end if;
            end if;
        end if;
	-- write-enable logic for the KCMs, but in the Host clock domain.
	-- It's up to the FDBK_LOOP module to move these signals to the
	-- 40 MHz signal processing domain in which the KCMs live.
        if (host_write_cycle and ADDR_reg(13 downto 2) = "000000000011") then
            KCM_load_cmd <= '1';
        else
            KCM_load_cmd <= '0';
        end if;

	-- write-enable logic for the serial port.  Just like the KCM write
	-- strobes above, but no problems with clock domain since the serial
	-- port runs on the host clock. 
        if((WE_host = '0') and (CS0_host = '0') and (ADDR_reg(13 downto 2) = "000000000110")) then
            host_sport_write <= '1';
        else
            host_sport_write <= '0';
        end if;

	-- write-enable logic for the bank of timer registers.
	-- This decoding matches for 0x0080 through 0x00bc, of which the 8 words at
	-- 0x0080 through 0x09c are actually used.
        if((WE_host = '0') and (CS0_host = '0') and (ADDR_reg(13 downto 4) = "0000000010")) then
	    host_timer_write <= '1';
        else
	    host_timer_write <= '0';
        end if;

	-- write-enable logic for the feedforward table.  This goes into one half of
	-- a dual port RAM.  As long as writes don't take place during the RF pulse,
	-- there's no chance of using corrupted data for the feedforward.  Maybe put
	-- an error bit on this coincidence?
	-- Enable on addresses 0x3000 through 0x3ffc, although only 512 elements
	-- are implemented, so beyond 0x3000 through 0x37fc the address wraps.
        if((WE_host = '0') and (CS0_host = '0') and (ADDR_reg(13 downto 10) = "0011")) then
            feedforward_we <= '1';
        else
            feedforward_we <= '0';
        end if;
    
	-- reading the status register (0x0008) causes error bits to clear.
	-- Latch the status bits in a register so we can se what the
	-- bits were before the clear happened. 
        if (RD_host = '0' and CS0_host = '0' and ADDR_REG = "00000000000010") then
    	    error_clear_cmd <= '1';
        else
            error_clear_cmd <= '0';
--            status <= RF_ON_sync & SYNC_sync & sync_error & sync_missing &
--	              ext_trigger_skipped_latch & KCM_load_busy & clk40_missing & urst &
--                      gray_sample & gray_previous &
--                      daov_latch & dbov_latch & dcov_latch & ddov_latch;
	    status <= RF_ON_sync & SYNC_sync & KCM_load_busy & urst & gray_sample & gray_previous;
	    errors <= sync_error & sync_missing & ext_trigger_skipped_latch & clk40_missing &
	              daov_latch & dbov_latch & dcov_latch & ddov_latch;
        end if;

    -- nearly binary tree to select PD output to host
    if (RD_host = '0' and CS0_host = '0') then
      if (ADDR_reg(13 downto 11) = "000") then          -- 0x0000 through 0x1ffc
        if (ADDR_reg(10 downto 2) = "000000000") then   -- 0x0000 through 0x000c
          if (ADDR_reg(1) = '0') then                   -- 0x0000 through 0x0004
            if (ADDR_reg(0) = '0') then			-- 0x0000
	      PD <= "0111011100000001";			-- product code 0x7701
            else                                  	-- 0x0004
              PD <= "0000000000010001";			-- firmware version #17
	    end if;
	  else						-- 0x0008 through 0x001c
	    if (ADDR_reg(0) = '0') then                 -- 0x0008 (designed)
	    -- status bits, could add more decoding here too
	      PD <= "00000000" & status;
	    else                                        -- 0x000c (designed)
	      PD <= "00000000" & errors;
	    end if;
	  end if;
	else      					-- 0x0010 through 0x1ffc
	  if (ADDR_reg(10 downto 4) = "0000000") then   -- 0x0010 through 0x003c
	    if (ADDR_reg(3) = '0') then                 -- 0x0010 through 0x001c
	      PD <= address_at_pulse_end;		-- 0x0014 (designed)
	    else					-- 0x0020 through 0x003c
              PD <= conv_std_logic_vector(ARG => SET_I, SIZE => 14) & "00";
	    end if;
	  else                                          -- 0x0040 through 0x1ffc
	    if (ADDR_reg(10 downto 5) = "000000") then	-- 0x0040 through 0x007c
	      PD <= host_sport_odata;
	    else					-- 0x0080 through 0x1ffc
	      PD <= host_timer_odata;
	    end if;
	  end if;
	end if;
      else                                      	-- 0x2000 through 0xfffc
      	if (ADDR_reg(13) = '0') then		        -- 0x2000 through 0x7ffc
      	  if (ADDR_reg(12) = '0') then                  -- 0x2000 through 0x3ffc
      	    PD <= "00000000" & feedforward_host_dout;	-- 0x3000 through 0x3ffc allocated,
      	  						-- 0x3000 through 0x37fc implemented
      	  else						-- 0x4000 through 0x7ffc
      	      -- Don't sign extend the data for buffer readout,
      	      -- they are 0-4095 offset binary values direct from the ADC.
            PD <=   "0000" & decay_TRACE_dout;          -- read out the decay_trace_buffer
          end if;
      	else						-- 0x8000 through 0xfffc
      	  if (ADDR_reg(12) = '0') then			-- 0x8000 through 0xbffc
      	    if (ADDR_reg(11) = '0') then		-- 0x8000 through 0x9ffc
      	      PD <= "0000" & wide_trace_dout(31 downto 21) & "0";
      	    else					-- 0xa000 through 0xbffc
      	      PD <= "0000" & wide_trace_dout(20 downto 10) & "0";
      	    end if;
      	  else						-- 0xc000 through 0xffff
      	      PD <= "0000" & wide_trace_dout( 9 downto  0) & "00";      	      
          end if;
        end if;
      end if;
    else
      PD <= "ZZZZZZZZZZZZZZZZ";			-- high impedance unless actively reading
    end if;
  end if;
end process host_IF;

-- DAC controls
OCLK <= not CLK_40MHz;
OSEL <= sync_count(0) xor iq_hand_flip;
SLEEP <= LO;


end DESIGN_TOP_arch;
