# first arg, data width
$n=16;
if ($ARGV[0]>1 && $ARGV[0]<200) { $n=int($ARGV[0]+.5); }
$n1=$n-1;

# second arg, address bits
$o=10;
if ($ARGV[1]>=8 && $ARGV[1]<=12) { $o=int($ARGV[1]+.5); }
$o1=$o-1;
# capacity in words
$s=1<<$o;
# configuration width of each 4K block RAM
$bw=4096/$s;
$bw1=$bw-1;
$bram="RAMB4_S${bw}_S${bw}";    # component I use
$aggram="RAMDP${s}X${n}";       # entity I define

# can't handle fractional block width
if (int($n/$bw)*$bw != $n) {
   die "Can't handle fractional block width";
}

$howmany = $n/$bw;

print "-- automatically generated from perl bram.pl $n $o
-- Larry Doolittle   <ldoolitt\@recycle.lbl.gov>  March 2002

-- $s x $n dual port ram, aggregated from
-- $howmany instances of $s x $bw Virtex block ram

-- No, I don't trust myself to use VHDL's generate construct, and I
-- will need to take this approach anyway when I move to Verilog.

library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;

entity $aggram is
  port (
	ADDRa: IN std_logic_vector($o1 downto 0);
	RSTa : IN std_logic;
	ENa  : IN std_logic;
	WEa  : IN std_logic;
	CLKa : IN std_logic;
 	DIa  : IN  std_logic_vector($n1 downto 0);
	DOa  : OUT std_logic_vector($n1 downto 0);

	ADDRb: IN std_logic_vector($o1 downto 0);
	RSTb : IN std_logic;
	ENb  : IN std_logic;
	WEb  : IN std_logic;
	CLKb : IN std_logic;
 	DIb  : IN  std_logic_vector($n1 downto 0);
	DOb  : OUT std_logic_vector($n1 downto 0)
  );
end $aggram;

architecture virtex of $aggram is

component $bram is
  port (
	ADDRa: IN std_logic_vector($o1 downto 0);
	RSTa : IN std_logic;
	ENa  : IN std_logic;
	WEa  : IN std_logic;
	CLKa : IN std_logic;
 	DIa  : IN  std_logic_vector($bw1 downto 0);
	DOa  : OUT std_logic_vector($bw1 downto 0);

	ADDRb: IN std_logic_vector($o1 downto 0);
	RSTb : IN std_logic;
	ENb  : IN std_logic;
	WEb  : IN std_logic;
	CLKb : IN std_logic;
 	DIb  : IN  std_logic_vector($bw1 downto 0);
	DOb  : OUT std_logic_vector($bw1 downto 0)
  );
end component;

begin
";

for ($i=0; $i*$bw<$n; $i++) {
   $y= $i * $bw;
   $x= $y + $bw - 1;
   print "
b$i: $bram port map (
	ADDRa => ADDRa,
	RSTa  => RSTa,
	ENa   => ENa,
	WEa   => WEa,
	CLKa  => CLKa,
	DIa   => DIa($x downto $y),
	DOa   => DOa($x downto $y),

	ADDRb => ADDRb,
	RSTb  => RSTb,
	ENb   => ENb,
	WEb   => WEb,
	CLKb  => CLKb,
	DIb   => DIb($x downto $y),
	DOb   => DOb($x downto $y)
);
";}

print "
end virtex;
";
