library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
--use IEEE.numeric_std.all;

entity sport is
  port (
  	write: in std_logic;
  	addr : in std_logic_vector (1 downto 0);
  	idata: in std_logic_vector (15 downto 0);
  	din  : in std_logic;
   	clk  : in std_logic;
   	dout : out std_logic;
  	odata: out std_logic_vector (15 downto 0);
  	sclk : out std_logic;
  	cs0  : out std_logic;
  	cs1  : out std_logic;
  	cs2  : out std_logic;
  	cs3  : out std_logic
       );
end sport;
  	
architecture sport_arch of sport is
signal ldin    : std_logic;
signal cke     : std_logic;
signal slow    : std_logic;
signal stop    : std_logic;
signal typ     : std_logic;
signal running : std_logic;
signal cnt     : std_logic_vector (8 downto 0);
signal data    : std_logic_vector (15 downto 0);

signal writeORslow : std_logic;
signal writeORstop : std_logic;
signal slow_cnt16  : std_logic;
signal slow_cnt24  : std_logic;
signal stop_cnt17  : std_logic;
signal stop_cnt25  : std_logic;

begin

sclk <= cnt(3);
cke <= '1' when (cnt(3 downto 0) = "1111") else '0';
slow_cnt16 <= '1' when (cnt(8 downto 4) = "01111") else '0';
slow_cnt24 <= '1' when (cnt(8 downto 4) = "10111") else '0';
stop_cnt17 <= '1' when (cnt(8 downto 4) = "10000") else '0';
stop_cnt25 <= '1' when (cnt(8 downto 4) = "11000") else '0';         
dout <= data(15);
odata <= data(15 downto 1) & running;

slow <= cke and slow_cnt24 when (typ = '1') else
	cke and slow_cnt16;
stop <= cke and stop_cnt25 when (typ = '1') else
	cke and stop_cnt17;	

writeORslow <= write or slow;
writeORstop <= write or stop;
	
u0: process (clk)
begin
  if (clk'event and clk = '1') then
    if (writeORstop = '1') then
      running <= write;
    end if;
    if (writeORslow = '1') then
      cs0 <= not (write and (not addr(1)) and (not addr(0)));
      cs1 <= not (write and (not addr(1)) and (    addr(0)));
      cs2 <= not (write and (    addr(1)) and (not addr(0)));
      cs3 <= not (write and (    addr(1)) and (    addr(0)));
    end if;
    if (write = '1') then
      typ <= addr(1);
    end if;
    if (cnt(3 downto 0) = "0111") then
      ldin <= din;
    end if;
  end if;
end process u0;

u1: process (clk)
begin 
  if (clk'event and clk = '1') then
    if (running = '1') then
      if (stop = '1') then
        cnt <= (others => '0');
      else
        cnt <= cnt + 1;
      end if;
    end if;
  end if;
end process u1;

u2: process (clk)
begin
  if (clk'event and clk = '1') then
    if (running = '1') then
      if (cke = '1') then
        data <= data(14 downto 0) & ldin;
      end if;
    elsif (write = '1') then
      data <= idata;
    end if;
  end if;
end process u2;

end sport_arch;
--if (running & cke) then data <= data(14 downto 0) & ldin;
--elsif (write) then data <= idata;

