#!/usr/bin/bltwish

global lfont
set lfont {Helvetica 10}
if { [info commands "namespace"] == "namespace" } {
    if { $tcl_version >= 8.0 } {
        namespace import blt::*
    } else {
        catch { import add blt }
    }
    if { $tcl_version >= 8.0 } {
        namespace import -force blt::tile::*
    } else {
        import add blt::tile
    }
}

proc trace_create { base name color } {
	foreach n { 0 1 } {
		set s [ lindex {r i} $n ]
		set y $name${s}Vec
		vector ${y}(256)
		$y set { 0 }
		$base element create $name$n -xdata tVec -ydata $y \
			-mapx time -mapy bits -color [lindex $color $n] \
			-pixels 3 -linewidth 0
	}
}

proc create_llrf_graph { base } {
	# Create a new graph
	graph $base.g -plotbackground grey50 -width 600 -height 450 -title "RF Pulse"

	$base.g axis create time -title "Time (us)" -min 0 -max 256
	$base.g xaxis use time
	$base.g axis create bits -title "ADC bits" -min -905 -max 905
	$base.g yaxis use bits

	# Create two vectors and add them to the graph. 
	vector tVec(256)
	tVec seq 0 255 1

	trace_create $base.g fwd { blue blue3 }
	trace_create $base.g rev { green green3 }
	trace_create $base.g cav { orange orange3 }
	vector txVec(1024)
	txVec set { 0 }
	vector txTime(1024)
	txTime seq 0 255.75 0.25
	$base.g element create tx -xdata txTime -ydata txVec \
		-mapx time -mapy bits -color red -pixels 3 -linewidth 0
}

# disabled, out of date
proc fill_graph { fname } {
	set f [open $fname r]
	set fvr { }
	set fvi { }
	set rvr { }
	set rvi { }
	set cvr { }
	set cvi { }
	set xv { }
	while {[gets $f line] >= 0} {
		# do stuff with $line
		set t [lindex $line 0]
		set m [lindex {r i} [expr "$t%2"]]
		set s [expr "2*(($t/2)%2)-1"]
		lappend fv$m [expr $s*[lindex $line 1]]
		lappend rv$m [expr $s*[lindex $line 2]]
		lappend cv$m [expr $s*[lindex $line 3]]
		lappend xv$m [expr $s*[lindex $line 4]]
	}
	fwdrVec set $fvr
	fwdiVec set $fvi
	revrVec set $rvr
	reviVec set $rvi
	cavrVec set $cvi
	caviVec set $cvi
	# revVec set $rv
	# cavVec set $cv
	# t0Vec set $tv0
	# t1Vec set $tv1
	# t2Vec set $tv2
	# t3Vec set $tv3
	close $f
}

proc new_val { name1 name2 op } {
	global ho $name1 ${name1}_scale ${name1}_add
	set a [ expr "\$${name1}_add" ]
	# set v [ expr "int(\$$name1*\$${name1}_scale)" ]
	set v [ expr "\$$name1*\$${name1}_scale" ]
	puts $ho "$a $v $name1"
	flush $ho
}

proc llrf_text { w var add def } {
	global $var ${var}_scale ${var}_add lfont
	set ${var}_scale 1
	set ${var}_add $add
	trace variable $var w new_val
	set $var $def
	frame $w.$var
	entry $w.$var.v -textvariable $var -width 8
	label $w.$var.l -text $var -width 16 -justify right -font $lfont
	pack $w.$var.v $w.$var.l -side right
	pack $w.$var
}

proc llrf_slider { w var add def min max scale unit } {
	global $var ${var}_scale ${var}_add lfont
	set ${var}_scale $scale
	set ${var}_add   $add
	trace variable $var w new_val
	set $var $def
	frame $w.$var
	scale $w.$var.s -orient horizontal -variable $var \
	                -from $min -to $max -length 200 -width 9 -font $lfont
	label $w.$var.l -text $var -width 16 -justify right
	label $w.$var.u -text $unit -width 7 -justify left
	pack $w.$var.u $w.$var.s $w.$var.l -side right
	pack $w.$var
}

proc monitor_var { w var unit } {
	global $var lfont
	frame $w.$var
	label $w.$var.l -text $var -font $lfont
	label $w.$var.v -textvariable $var -width 15 -font $lfont
	label $w.$var.u -text $unit -font $lfont
	pack $w.$var.u $w.$var.v $w.$var.l -side right
}

global ho hi
if { $argc == 2 } {
	set host [lindex $argv 0]
	set port [lindex $argv 1]
	puts "Commands to network socket $host:$port"
	set ho [ socket -async $host $port ]
	set hi $ho
} elseif { $argc == 1 } {
	set filename [lindex $argv 0]
	puts "Commands to file $filename"
	set ho [ open $filename w ]
	set hi ""
} else {
	puts "Commands to stdout"
	fconfigure file0 -blocking 0
	set ho file1
	# set hi file0 if you want to ack from the keyboard
	set hi file0
}

proc handle_waveform { foo cmd } {
	# puts "# hello world waveform $foo" 
	${foo}Vec set [ lrange $cmd 1 end ]
}

proc bitbox_color { name word } {
	global ${name}frame ${name}bits
	set s [expr \$${name}frame]
	set w $word
	foreach l [expr \$${name}bits] {
		if { [expr "$w%2"] } { $s.$l configure -bg red } else {$s.$l configure -bg green}
		set w [expr "int($w/2)"]
	}
}

proc do_cmd_er { cmd } {
	bitbox_color errs [lindex $cmd 1]
}

proc do_cmd_st { cmd } {
	bitbox_color stat [lindex $cmd 1]
}

proc do_cmd_bw { cmd } {
	global monitor_bw
	set monitor_bw [lindex $cmd 1]
}

proc do_cmd_df { cmd } {
	global monitor_df
	set monitor_df [lindex $cmd 1]
}

proc do_cmd_ds { cmd } {
	global monitor_ds
	set monitor_ds [lindex $cmd 1]
}

proc do_cmd_pl { cmd } {
	global monitor_pl
	set monitor_pl [lindex $cmd 1]
}

proc do_cmd_cy { cmd } {
	global monitor_cy
	set monitor_cy [lindex $cmd 1]
}
	
proc do_cmd_ar { cmd } {
	handle_waveform fwdr $cmd
}

proc do_cmd_ai { cmd } {
	handle_waveform fwdi $cmd
}

proc do_cmd_br { cmd } {
	handle_waveform revr $cmd
}

proc do_cmd_bi { cmd } {
	handle_waveform revi $cmd
}

proc do_cmd_cr { cmd } {
	handle_waveform cavr $cmd
}

proc do_cmd_ci { cmd } {
	handle_waveform cavi $cmd
}

proc do_cmd_tx { cmd } {
	handle_waveform tx $cmd
}

proc process_command { } {
	global hi
	while {[gets $hi cmd] >= 0} {
		if { [ eof $hi ] } {
			# what goes here?
			return
		}
		# puts "processcommand: $cmd"
		set pcode [lindex $cmd 0]
		if [ catch { do_cmd_$pcode $cmd } ] {
			# puts "# couldn't $pcode"
		}
		return
	}
}

proc bitsel { name b } {
	global ${name}_${b} $name
	set v [expr \$${name}_${b}]
	# puts "$name $b $v"
	set tmp [expr \$$name]
	if { $v } {
		set tmp [expr "$tmp | (1<<$b)"]
	} else {
		set tmp [expr "$tmp & ~(1<<$b)"]
	}
	# puts "$name $b $v $tmp"
	set $name $tmp
}

proc cfgbox { base name add def bits} {
	global $name ${name}_add ${name}_scale
	set ${name}_add $add
	set ${name}_scale 1
	trace variable $name w new_val
	set $name $def
	frame $base.$name
	set b 0
	set s $def
	foreach l $bits {
		if {$l!="--"} {
		global ${name}_${b}
		set ${name}_${b} [expr $s%2]
		checkbutton $base.$name.$l -text $l -font {Helvetica 10} \
			-variable ${name}_${b} -command "bitsel $name $b"
		pack $base.$name.$l
		}
		set b [expr $b+1]
		set s [expr int($s/2)]
	}
	pack $base.$name -side left
}

proc bitbox { base name bits } {
	global ${name}frame ${name}bits
	set f $base.$name
	frame $f
	set ${name}bits $bits
	foreach l $bits {
		label $f.$l -text $l -width 7 -bg green
		pack $f.$l -side bottom
	}
	set ${name}frame $base.$name
}

# No other convenient point for these, might as well be here
puts $ho "ss 0 1.0"

frame .top_half
frame .top_half.l
cfgbox .top_half.l conf "cf 0" 7937 {te ps -- cf ir sr ha ts0 ts1 fs is tp iq}
bitbox .top_half.l stat {grayl0 grayl1 grays0 grays1 urst kcmb sync rfon}
button .top_half.l.quit -text Quit -command exit
pack .top_half.l.conf -side left
pack .top_half.l.quit .top_half.l.stat
create_llrf_graph .top_half
bitbox .top_half errs {ddov dcov dbov daov miss40 trigs syncm synce} 
pack .top_half.l -side left
pack .top_half.g -side left
pack .top_half.errs -side left
pack .top_half

fill_graph "dtest3.dat"

frame .middle
monitor_var .middle monitor_ds ""
monitor_var .middle monitor_pl ""
monitor_var .middle monitor_cy ""
pack .middle.monitor_ds .middle.monitor_pl .middle.monitor_cy -side left
pack .middle
frame .lslide
monitor_var .lslide monitor_df kHz
monitor_var .lslide monitor_bw kHz
pack .lslide.monitor_df .lslide.monitor_bw
llrf_text   .lslide decimate    "td 0"     3
llrf_text   .lslide d_ignore    "di 0"    40
llrf_text   .lslide d_length    "dl 0"   600
# llrf_text .lslide config      "cf 0"     0
llrf_slider .lslide mset        "sm 0"     0       0    100  .01  "%"
llrf_slider .lslide aset        "sp 0"     0    -200    200    1  degrees
llrf_slider .lslide mgain       "gm 0"     0       0    100  .01  "%"
llrf_slider .lslide again       "gp 0"   +48    -200    200    1  degrees
llrf_slider .lslide intr        "ig 0"   650   -2048   2047   16  bits
pack .lslide -side left
frame .rslide
llrf_slider .rslide t_idle      "tm 0"   100       0   2000   10  us
llrf_slider .rslide t_adc_warm  "tm 2"    25       0    200   10  us
llrf_slider .rslide t_feedfwd   "tm 3"    15       0    200   10  us
llrf_slider .rslide t_feedbk    "tm 4"    65       0   2000   10  us
llrf_slider .rslide t_decay     "tm 5"    22       0    200   10  us
llrf_slider .rslide t_inhibit   "tm 1" 11000       0  15000  2.5  us
llrf_slider .rslide t_wait      "tm 7"  8000       0   9000  2.5  us
pack .rslide -side left

update
flush $ho
if { $hi != "" } {
    fileevent $hi readable {
        process_command
    }
}
puts $ho "oe 0 32"
flush $ho
