#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/time.h>

/* #define TESTME */
#define DBG(x)

static inline int limul(int a, int b)
{
	long long res;
	res = (long long)a * b;
	return res>>31;
}

static inline int liinv(int x, int *s)
{
	long long y;
	int i=0, n=0;
	y = (long long)INT_MAX*INT_MAX/x;
	if (y<0) { y=-y; n=1;}
	while (y>INT_MAX) { i++; y=y>>1; }
	if (n) y=-y;
	*s=i;
	return y;
}

static int decay(int n, short d[], int r[])
{
	int i;
	int d_i1, d_i2, d_i3, y;
	int T1, T2, S11, S12, S22;
	/* double det, A1, A2; */
	int iA1, iA2;
	
	T1=0; T2=0; S11=0; S12=0; S22=0;
#define S_SCALE 2
	d_i1 = d[0];
	d_i2 = d[1];
	for (i=0; i<n-2; i++) {
		d_i3 = d[i+2];
		y = d_i1+d_i3;
		/* printf("%2d %5d %5d %5d %5d\n",i,d_i1,d_i2,d_i3,y); */
		T1  += (y*d_i1)    >> S_SCALE;
		T2  += (y*d_i2)    >> S_SCALE;
		S11 += (d_i1*d_i1) >> S_SCALE;
		S12 += (d_i1*d_i2) >> S_SCALE;
		S22 += (d_i2*d_i2) >> S_SCALE;
		d_i1=d_i2;
		d_i2=d_i3;
	}
#if 0
	{
		double rS11=S11, rS12=S12, rS22=S22, rT1=T1, rT2=T2;
		double A1, A2, det;
		/* printf("%f %f %f %f %f\n",rT1,rT2,rS11,rS12,rS22); */
		det = (rS11*rS22-rS12*rS12);
		A1 = ( rS22*rT1-rS12*rT2)/det;
		A2 = ( rS11*rT2-rS12*rT1)/det;
		/* printf("A1=%f A2=%f\n",A1,A2); */
		iA1 = floor(A1*INT_MAX+0.5);
		iA2 = floor(A2*INT_MAX+0.5);
	}
#else
	{
		int i, a, b=0, det, det_inv;
		DBG(printf("%d %d %d %d %d\n",T1,T2,S11,S12,S22);)
		b=abs(S11);
		a=abs(S12); if (a>b) b=a;
		a=abs(S22); if (a>b) b=a;
		a=abs(T1 ); if (a>b) b=a;
		a=abs(T2 ); if (a>b) b=a;
		i=0;
		while (b<INT_MAX/4) { i++; b=b<<1; }
		DBG(printf("shift by %d\n",i);)
		S11 = S11<<i;
		S12 = S12<<i;
		S22 = S22<<i;
		T1  = T1 <<i;
		T2  = T2 <<i;
		det =  limul(S11,S22)-limul(S12,S12);
		if (abs(det) < (1<27)) return 4;   /* error code, determinant
			should not be small difference of large numbers */
		det_inv = liinv(det,&i);
		DBG(printf("det %d  det_inv %d  scale %d\n",det,det_inv,i);)
		iA1 = limul( limul(S22,T1 )-limul(S12,T2 ),det_inv)<<i;
		iA2 = limul( limul(S11,T2 )-limul(S12,T1 ),det_inv)<<i;
	}
#endif
	r[0]=iA1;
	r[1]=iA2;
	return 0;
}

static int post_proc(int r[], double dt, double p[])
{
	int iA1=r[0], iA2=r[1];
	double c, d, a, b;
	double A1=iA1/(double)INT_MAX;
	double A2=iA2/(double)INT_MAX;
	c=A2/2;
	d=sqrt(1-A1-c*c);
#if 0
	a=-log(1-A1)/2;
	b=atan2(d,c);
	b-=M_PI/2;   // fake, because we don't want to see 10 MHz offset
#else
	{
		double x;
		x=c/d;
		b=-x*(1-x*x/3.0);
		x=A1;
		a=x*(0.5+x*(0.25+x*(0.16666666666+x*0.125)));
	}
#endif
#if 0
	printf("iA1=%d iA2=%d\n",iA1,iA2);
	printf("%.6f %.6f %.6f %.6f %.6f %.6f\n",A1,A2,c,d,a,b);
#endif
	p[0]=b/dt/2/M_PI/1000;  // df, cavity detune in kHz
	p[1]=a/dt/2/M_PI/1000;  // bw, cavity bandwidth in kHz
	return 0;
}

int analyze_decay(int npt, short data[], double dresult[])
{
	double dt=25e-9;            // seconds for 40 MHz sample rate
	int i, stat, iresult[2], sum;
	if (npt<16) return 1;
	/* printf("# data start:"); */
	/* for (       i=0;     i<8;   i++) printf(" %d",data[i]); */
	for (sum=0, i=0;     i<8;   i++) sum+=data[i]*data[i];
	DBG(printf("start: sum %d\n",sum);)
	if (sum<100000) return 2;
	/* printf("# data end:"); */
	/* for (       i=npt-8; i<npt; i++) printf(" %d",data[i]); */
	for (sum=0, i=npt-8; i<npt; i++) sum+=data[i]*data[i];
	DBG(printf("end: sum %d\n",sum);)
	if (sum>70000  ) return 3;

	/* take out the ADC offset */
	for (sum=0, i=npt-8; i<npt; i++) sum+=data[i];
	DBG(printf("average = %d/8\n",sum);)
	sum=sum>>3;
	for (i=0; i<npt; i++) data[i]-=sum;
	

	stat = decay(npt, data, iresult);
	if (stat) return stat;
	stat = post_proc(iresult, dt, dresult);
	return stat;
}

#ifdef TESTME
int main(int argc, char *argv[])
{
	short *d;
	int i, j, n, rd, rcode;
	/* int r[2]; */
	int diff;
	double p[2];
	struct timeval now, then;
	if (argc != 2) return 1;
	n = atoi(argv[1]);
	d = malloc(n*sizeof *d);
	if (d==NULL) return 2;
	for (i=0; i<n; i++) {
		if (scanf("%d\n",&rd)!=1) return 1;
		d[i]=rd;
	}

	gettimeofday(&then,NULL);
	for (j=0; j<1; j++) {
		/* decay(n,d,r); */
		rcode=analyze_decay(n,d,p);
	}
	/* post_proc(r,25.0e-9,p); */
	gettimeofday(&now,NULL);
	diff=(now.tv_sec-then.tv_sec)*1000000;
	diff+=(now.tv_usec-then.tv_usec);
	fprintf(stderr,"time %d us\n",diff);
	{ double Q;
	Q=402.5e3/p[1];
	printf("rcode = %d df = %.3f kHz   BW = %.3f kHz   Q = %.1f\n",
		rcode, p[0], p[1], Q);
	}
#if 0
	for (i=0; i<n; i++) {
		printf("%2d %5d\n",i,d[i]);
	}
#endif
	return 0;
}
#endif
