#include <stdio.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>          /* inet_ntoa */
#include <string.h>

#ifdef INET6
#define SOCKADDR sockaddr_storage
#define S_FAMILY __s_family
#define SERVER_AF AF_INET6
#else
#define SOCKADDR sockaddr_in
#define S_FAMILY sin_family
#define SERVER_AF AF_INET
#endif

/* Binds to the existing server_s, based on the configuration string
   in server_ip.  IPv6 version doesn't pay attention to server_ip yet.  */
int bind_server(int server_s, char *server_ip, int server_port)
{
#ifdef INET6
    struct sockaddr_in6 server_sockaddr;
    server_sockaddr.sin6_family = AF_INET6;
    memcpy(&server_sockaddr.sin6_addr, &in6addr_any, sizeof (in6addr_any));
    server_sockaddr.sin6_port = htons(server_port);
#else
    struct sockaddr_in server_sockaddr;
    memset(&server_sockaddr, 0, sizeof server_sockaddr);
#ifdef BSD                      /* uncomment for BSDs */
    /* NOTE:  This *will* get improved when the authors get a chance
       to learn what a good #define to use is
     */
    server_sockaddr.sin_len = sizeof server_sockaddr;
#endif
    server_sockaddr.sin_family = AF_INET;
    if (server_ip != NULL) {
        inet_aton(server_ip, &server_sockaddr.sin_addr);
    } else {
        server_sockaddr.sin_addr.s_addr = htonl(INADDR_ANY);
    }
    server_sockaddr.sin_port = htons(server_port);
#endif

    return bind(server_s, (struct sockaddr *) &server_sockaddr,
                sizeof (server_sockaddr));
}

FILE *setup_socket(int port)
{
	int fd;
	int one=1;
	static int server_s=0;
	struct SOCKADDR remote_addr; /* address */
	int remote_addrlen = sizeof (struct sockaddr_in);

	if (!server_s) {
		if ((server_s = socket(SERVER_AF, SOCK_STREAM, IPPROTO_TCP)) == -1) {
			perror("unable to create socket");
			exit(errno);
		}
		if (setsockopt(server_s, SOL_SOCKET, SO_REUSEADDR, &one, sizeof(one)) == -1) {
			perror("setsockopt:SO_REUSEADDR");
		}
		if (bind_server(server_s, NULL, port) == -1) {
			perror("unable to bind");
			exit(errno);
		}
		if (listen(server_s, 1) == -1) {
			perror("unable to listen");
			exit(errno);
		}

	}
    retry:
	fd = accept(server_s, (struct sockaddr *) &remote_addr,
		&remote_addrlen);
	if (fd == -1) {
		if (errno == EAGAIN) goto retry;
		perror("accept");
		exit(errno);
	}
	return fdopen(fd,"a+");
}
