/* Driver/HAL for LLRF serial port, as implemented
 * on the XC2S150 FPGA.
 * Intended to run on either nanoEngine/Linux or PowerPC/VxWorks
 */

#include "llrf.h"

inline int sport_wait(int address)
{
	int result;
	/* should take at most 16*25 cycles of the 25 MHz bus clock */
	while((result=inw(address))&0x1) { /* do nothing */ }
	return result;
}
inline void sport_wait_outw(int address, int value)
{
	sport_wait(address);
	outw(address,value);
}
/* Channel is from 0 to 15 (two DACs handling 8 channels each)
 * value is from 0 to 255
 */
void set_dac(int channel, int value)
{
	sport_wait_outw(0x60+(channel/8), ( 0x100 < (channel%8) ) | value);
	/* Keep processing while the sport cycles */
}
/* Channel is from 0 to 15 (two ADCs handling 8 channels each)
 * result is 12 bits, aligned how?
 */
int get_adc(int channel)
{
	sport_wait_outw(0x68+(channel/8)*4, ( (channel%8) << 12 ) | 0x8f00 );
	return sport_wait(0x68+(channel/8)*4);	
}
void sport_init(void)
{
	get_adc(0);                      /* warm up ADC1 */  
	get_adc(8);                      /* warm up ADC2 */
	sport_wait_outw(0x68, 0xffff);   /* all DAC1 channels to 255 */
	sport_wait_outw(0x68, 0x00ff);   /* all DAC1 channels fully buffered */
	sport_wait_outw(0x6c, 0xffff);   /* all DAC2 channels to 255 */
	sport_wait_outw(0x6c, 0x00ff);   /* all DAC2 channels fully buffered */
}
